@@ -3,6 +3,7 @@
  *  Copyright (C) 2008 - 2009 Erlend Hamberg <ehamberg@gmail.com>
  *  Copyright (C) 2009 Paul Gideon Dann <pdgiddie@gmail.com>
  *  Copyright (C) 2011 Svyatoslav Kuzmich <svatoslav1@gmail.com>
+ *  Copyright (C) 2012 - 2013 Simon St James <kdedevel@etotheipiplusone.com>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
@@ -24,7 +25,7 @@
 
 #include <QKeyEvent>
 #include <QString>
-#include <QCoreApplication>
+#include <QApplication>
 
 #include <kconfig.h>
 #include <kconfiggroup.h>
@@ -37,8 +38,15 @@
 #include "katevivisualmode.h"
 #include "katevireplacemode.h"
 #include "katevikeyparser.h"
+#include "katevikeymapper.h"
 #include "kateviemulatedcommandbar.h"
 
+using KTextEditor::Cursor;
+using KTextEditor::Document;
+using KTextEditor::Mark;
+using KTextEditor::MarkInterface;
+using KTextEditor::MovingCursor;
+
 KateViInputModeManager::KateViInputModeManager(KateView* view, KateViewInternal* viewInternal)
 {
   m_viNormalMode = new KateViNormalMode(this, view, viewInternal);
@@ -53,16 +61,25 @@ KateViInputModeManager::KateViInputModeManager(KateView* view, KateViewInternal*
 
   m_view->setCaretStyle( KateRenderer::Block, true );
 
-  m_replayingLastChange = false;
-  m_textualRepeat = false;
+  m_insideHandlingKeyPressCount = 0;
+
+  m_isReplayingLastChange = false;
+
+  m_isRecordingMacro = false;
+  m_macrosBeingReplayedCount = 0;
+  m_lastPlayedMacroRegister = QChar::Null;
+
+  m_keyMapperStack.push(QSharedPointer<KateViKeyMapper>(new KateViKeyMapper(this, m_view->doc(), m_view)));
 
   m_lastSearchBackwards = false;
+  m_lastSearchCaseSensitive = false;
+  m_lastSearchPlacedCursorAtEndOfMatch = false;
 
   jump_list = new QList<KateViJump>;
   current_jump = jump_list->begin();
   m_temporaryNormalMode = false;
 
-  m_mark_set_inside_viinputmodemanager = false;
+  m_markSetInsideViInputModeManager = false;
 
   connect(m_view->doc(),
            SIGNAL(markChanged(KTextEditor::Document*, KTextEditor::Mark,
@@ -74,6 +91,13 @@ KateViInputModeManager::KateViInputModeManager(KateView* view, KateViewInternal*
   // KateViVisualMode (which inherits from KateViNormalMode) to respond
   // to changes in the document as well.
   m_viNormalMode->beginMonitoringDocumentChanges();
+
+  if (view->selection())
+  {
+    changeViMode(VisualMode);
+    m_view->setCursorPosition(Cursor(view->selectionRange().end().line(), view->selectionRange().end().column() - 1));
+    m_viVisualMode->updateSelection();
+  }
 }
 
 KateViInputModeManager::~KateViInputModeManager()
@@ -87,42 +111,56 @@ KateViInputModeManager::~KateViInputModeManager()
 
 bool KateViInputModeManager::handleKeypress(const QKeyEvent *e)
 {
-  bool res;
+  m_insideHandlingKeyPressCount++;
+  bool res = false;
+  bool keyIsPartOfMapping = false;
+  const bool isSyntheticSearchCompletedKeyPress = m_view->viModeEmulatedCommandBar()->isSendingSyntheticSearchCompletedKeypress();
 
-  if (m_view->viModeEmulatedCommandBar()->isVisible())
+  // With macros, we want to record the keypresses *before* they are mapped, but if they end up *not* being part
+  // of a mapping, we don't want to record them when they are played back by m_keyMapper, hence
+  // the "!isPlayingBackRejectedKeys()". And obviously, since we're recording keys before they are mapped, we don't
+  // want to also record the executed mapping, as when we replayed the macro, we'd get duplication!
+  if (isRecordingMacro() && !isReplayingMacro() && !isSyntheticSearchCompletedKeyPress && !keyMapper()->isExecutingMapping() && !keyMapper()->isPlayingBackRejectedKeys())
   {
-    res = m_view->viModeEmulatedCommandBar()->handleKeyPress(e);
+    QKeyEvent copy( e->type(), e->key(), e->modifiers(), e->text() );
+    m_currentMacroKeyEventsLog.append(copy);
   }
-  else
+
+  if (!isReplayingLastChange() && !isSyntheticSearchCompletedKeyPress)
   {
-    // record key press so that it can be repeated via "."
-    if (!isReplayingLastChange()) {
+    // Hand off to the key mapper, and decide if this key is part of a mapping.
+    if (e->key() != Qt::Key_Control && e->key() != Qt::Key_Shift && e->key() != Qt::Key_Alt && e->key() != Qt::Key_Meta)
+    {
+      const QChar key = KateViKeyParser::self()->KeyEventToQChar(*e);
+      if (keyMapper()->handleKeypress(key))
+      {
+        keyIsPartOfMapping = true;
+        res = true;
+      }
+    }
+  }
+
+  if (!keyIsPartOfMapping)
+  {
+    if (!isReplayingLastChange() && !isSyntheticSearchCompletedKeyPress) {
+      // record key press so that it can be repeated via "."
       QKeyEvent copy( e->type(), e->key(), e->modifiers(), e->text() );
       appendKeyEventToLog( copy );
     }
-    // FIXME: I think we're making things difficult for ourselves here.  Maybe some
-    //        more thought needs to go into the inheritance hierarchy.
-    switch(m_currentViMode) {
-      case NormalMode:
-        res = m_viNormalMode->handleKeypress(e);
-        break;
-      case InsertMode:
-        res = m_viInsertMode->handleKeypress(e);
-        break;
-      case VisualMode:
-      case VisualLineMode:
-      case VisualBlockMode:
-        res = m_viVisualMode->handleKeypress(e);
-        break;
-      case ReplaceMode:
-        res = m_viReplaceMode->handleKeypress(e);
-        break;
-      default:
-        kDebug( 13070 ) << "WARNING: Unhandled keypress";
-        res = false;
+
+    if (m_view->viModeEmulatedCommandBar()->isActive())
+    {
+      res = m_view->viModeEmulatedCommandBar()->handleKeyPress(e);
+    }
+    else
+    {
+      res = getCurrentViModeHandler()->handleKeypress(e);
     }
   }
 
+  m_insideHandlingKeyPressCount--;
+  Q_ASSERT(m_insideHandlingKeyPressCount >= 0);
+
   return res;
 }
 
@@ -132,7 +170,6 @@ void KateViInputModeManager::feedKeyPresses(const QString &keyPresses) const
   Qt::KeyboardModifiers mods;
   QString text;
 
-  kDebug( 13070 ) << "Repeating change";
   foreach(const QChar &c, keyPresses) {
     QString decoded = KateViKeyParser::self()->decodeKeySequence(QString(c));
     key = -1;
@@ -150,7 +187,7 @@ void KateViInputModeManager::feedKeyPresses(const QString &keyPresses) const
 
       // check if one or more modifier keys where used
       if (decoded.indexOf("s-") != -1 || decoded.indexOf("c-") != -1
-          || decoded.indexOf("m-") != -1 || decoded.indexOf("m-") != -1) {
+          || decoded.indexOf("m-") != -1 || decoded.indexOf("a-") != -1) {
 
         int s = decoded.indexOf("s-");
         if (s != -1) {
@@ -193,25 +230,52 @@ void KateViInputModeManager::feedKeyPresses(const QString &keyPresses) const
       text = decoded.at(0);
     }
 
+    // We have to be clever about which widget we dispatch to, as we can trigger
+    // shortcuts if we're not careful (even if Vim mode is configured to steal shortcuts).
     QKeyEvent k(QEvent::KeyPress, key, mods, text);
-
-    QCoreApplication::sendEvent(m_viewInternal, &k);
+    QWidget *destWidget = NULL;
+    if (QApplication::activePopupWidget())
+    {
+      // According to the docs, the activePopupWidget, if present, takes all events.
+      destWidget = QApplication::activePopupWidget();
+    }
+    else if (QApplication::focusWidget())
+    {
+      if (QApplication::focusWidget()->focusProxy())
+      {
+        destWidget = QApplication::focusWidget()->focusProxy();
+      }
+      else
+      {
+        destWidget = QApplication::focusWidget();
+      }
+    }
+    else
+    {
+      destWidget = m_view->focusProxy();
+    }
+    QApplication::sendEvent(destWidget, &k);
   }
 }
 
+bool KateViInputModeManager::isHandlingKeypress() const
+{
+  return m_insideHandlingKeyPressCount > 0;
+}
+
 void KateViInputModeManager::appendKeyEventToLog(const QKeyEvent &e)
 {
   if ( e.key() != Qt::Key_Shift && e.key() != Qt::Key_Control
       && e.key() != Qt::Key_Meta && e.key() != Qt::Key_Alt ) {
-    m_keyEventsLog.append(e);
+    m_currentChangeKeyEventsLog.append(e);
   }
 }
 
-void KateViInputModeManager::storeChangeCommand()
+void KateViInputModeManager::storeLastChangeCommand()
 {
   m_lastChange.clear();
 
-  QList<QKeyEvent> keyLog = isTextualRepeat() ? m_keyEventsBeforeInsert : m_keyEventsLog;
+  QList<QKeyEvent> keyLog = m_currentChangeKeyEventsLog;
 
   for (int i = 0; i < keyLog.size(); i++) {
     int keyCode = keyLog.at(i).key();
@@ -240,19 +304,114 @@ void KateViInputModeManager::storeChangeCommand()
 
     m_lastChange.append(key);
   }
-
-  if ( isTextualRepeat() ) {
-    // paste text from the insert register "^
-    m_lastChange.append( KateViKeyParser::self()->encodeKeySequence( "<esc>\"^p" ) );
-  }
-
+  m_lastChangeCompletionsLog = m_currentChangeCompletionsLog;
 }
 
 void KateViInputModeManager::repeatLastChange()
 {
-  m_replayingLastChange = true;
+  m_isReplayingLastChange = true;
+  m_nextLoggedLastChangeComplexIndex = 0;
   feedKeyPresses(m_lastChange);
-  m_replayingLastChange = false;
+  m_isReplayingLastChange = false;
+}
+
+void KateViInputModeManager::startRecordingMacro(QChar macroRegister)
+{
+  Q_ASSERT(!m_isRecordingMacro);
+  kDebug(13070) << "Recording macro: " << macroRegister;
+  m_isRecordingMacro = true;
+  m_recordingMacroRegister = macroRegister;
+  KateGlobal::self()->viInputModeGlobal()->clearMacro(macroRegister);
+  m_currentMacroKeyEventsLog.clear();
+  m_currentMacroCompletionsLog.clear();
+}
+
+void KateViInputModeManager::finishRecordingMacro()
+{
+  Q_ASSERT(m_isRecordingMacro);
+  m_isRecordingMacro = false;
+  KateGlobal::self()->viInputModeGlobal()->storeMacro(m_recordingMacroRegister, m_currentMacroKeyEventsLog, m_currentMacroCompletionsLog);
+}
+
+bool KateViInputModeManager::isRecordingMacro()
+{
+  return m_isRecordingMacro;
+}
+
+void KateViInputModeManager::replayMacro(QChar macroRegister)
+{
+  if (macroRegister == '@')
+  {
+    macroRegister = m_lastPlayedMacroRegister;
+  }
+  m_lastPlayedMacroRegister = macroRegister;
+  kDebug(13070) << "Replaying macro: " << macroRegister;
+  const QString macroAsFeedableKeypresses = KateGlobal::self()->viInputModeGlobal()->getMacro(macroRegister);
+  kDebug(13070) << "macroAsFeedableKeypresses:  " << macroAsFeedableKeypresses;
+
+  m_macrosBeingReplayedCount++;
+  m_nextLoggedMacroCompletionIndex.push(0);
+  m_macroCompletionsToReplay.push(KateGlobal::self()->viInputModeGlobal()->getMacroCompletions(macroRegister));
+  m_keyMapperStack.push(QSharedPointer<KateViKeyMapper>(new KateViKeyMapper(this, m_view->doc(), m_view)));
+  feedKeyPresses(macroAsFeedableKeypresses);
+  m_keyMapperStack.pop();
+  m_macroCompletionsToReplay.pop();
+  m_nextLoggedMacroCompletionIndex.pop();
+  m_macrosBeingReplayedCount--;
+  kDebug(13070) << "Finished replaying: " << macroRegister;
+}
+
+bool KateViInputModeManager::isReplayingMacro()
+{
+  return m_macrosBeingReplayedCount > 0;
+}
+
+void KateViInputModeManager::logCompletionEvent(const KateViInputModeManager::Completion& completion)
+{
+  // Ctrl-space is a special code that means: if you're replaying a macro, fetch and execute
+  // the next logged completion.
+  QKeyEvent ctrlSpace( QKeyEvent::KeyPress, Qt::Key_Space, Qt::ControlModifier, " ");
+  if (isRecordingMacro())
+  {
+    m_currentMacroKeyEventsLog.append(ctrlSpace);
+    m_currentMacroCompletionsLog.append(completion);
+  }
+  m_currentChangeKeyEventsLog.append(ctrlSpace);
+  m_currentChangeCompletionsLog.append(completion);
+}
+
+KateViInputModeManager::Completion KateViInputModeManager::nextLoggedCompletion()
+{
+  Q_ASSERT(isReplayingLastChange() || isReplayingMacro());
+  if (isReplayingLastChange())
+  {
+    if (m_nextLoggedLastChangeComplexIndex >= m_lastChangeCompletionsLog.length())
+    {
+      kDebug(13070) << "Something wrong here: requesting more completions for last change than we actually have.  Returning dummy.";
+      return Completion("", false, Completion::PlainText);
+    }
+    return m_lastChangeCompletionsLog[m_nextLoggedLastChangeComplexIndex++];
+  }
+  else
+  {
+    if (m_nextLoggedMacroCompletionIndex.top() >= m_macroCompletionsToReplay.top().length())
+    {
+      kDebug(13070) << "Something wrong here: requesting more completions for macro than we actually have.  Returning dummy.";
+      return Completion("", false, Completion::PlainText);
+    }
+    return m_macroCompletionsToReplay.top()[m_nextLoggedMacroCompletionIndex.top()++];
+  }
+}
+
+void KateViInputModeManager::doNotLogCurrentKeypress()
+{
+  if (m_isRecordingMacro)
+  {
+    Q_ASSERT(!m_currentMacroKeyEventsLog.isEmpty());
+    m_currentMacroKeyEventsLog.pop_back();
+  }
+  Q_ASSERT(!m_currentChangeKeyEventsLog.isEmpty());
+  m_currentChangeKeyEventsLog.pop_back();
 }
 
 const QString KateViInputModeManager::getLastSearchPattern() const
@@ -288,6 +447,24 @@ ViMode KateViInputModeManager::getCurrentViMode() const
   return m_currentViMode;
 }
 
+KateViModeBase* KateViInputModeManager::getCurrentViModeHandler() const
+{
+  switch(m_currentViMode) {
+    case NormalMode:
+      return m_viNormalMode;
+    case InsertMode:
+      return m_viInsertMode;
+    case VisualMode:
+    case VisualLineMode:
+    case VisualBlockMode:
+      return m_viVisualMode;
+    case ReplaceMode:
+      return m_viReplaceMode;
+  }
+  kDebug( 13070 ) << "WARNING: Unknown Vi mode.";
+  return NULL;
+}
+
 void KateViInputModeManager::viEnterNormalMode()
 {
   bool moveCursorLeft = (m_currentViMode == InsertMode || m_currentViMode == ReplaceMode)
@@ -306,14 +483,13 @@ void KateViInputModeManager::viEnterNormalMode()
     addMark( m_view->doc(), '^', Cursor( m_view->cursorPosition() ), false, false );
   }
 
-  setTextualRepeat(false);
   changeViMode(NormalMode);
 
   if ( moveCursorLeft ) {
     m_viewInternal->cursorPrevChar();
   }
   m_view->setCaretStyle( KateRenderer::Block, true );
-  m_viewInternal->repaint ();
+  m_viewInternal->update ();
 }
 
 void KateViInputModeManager::viEnterInsertMode()
@@ -324,12 +500,11 @@ void KateViInputModeManager::viEnterInsertMode()
   {
     // Ensure the key log contains a request to re-enter Insert mode, else the keystrokes made
     // after returning from temporary normal mode will be treated as commands!
-    m_keyEventsLog.append(QKeyEvent(QEvent::KeyPress, QString("i")[0].unicode(), Qt::NoModifier, "i"));
+    m_currentChangeKeyEventsLog.append(QKeyEvent(QEvent::KeyPress, QString("i")[0].unicode(), Qt::NoModifier, "i"));
   }
-  m_keyEventsBeforeInsert = m_keyEventsLog;
   m_view->setCaretStyle( KateRenderer::Line, true );
   setTemporaryNormalMode(false);
-  m_viewInternal->repaint ();
+  m_viewInternal->update ();
 }
 
 void KateViInputModeManager::viEnterVisualMode( ViMode mode )
@@ -339,7 +514,7 @@ void KateViInputModeManager::viEnterVisualMode( ViMode mode )
   // If the selection is inclusive, the caret should be a block.
   // If the selection is exclusive, the caret should be a line.
   m_view->setCaretStyle( KateRenderer::Block, true );
-  m_viewInternal->repaint ();
+  m_viewInternal->update();
   getViVisualMode()->setVisualModeType( mode );
   getViVisualMode()->init();
 }
@@ -348,7 +523,7 @@ void KateViInputModeManager::viEnterReplaceMode()
 {
   changeViMode(ReplaceMode);
   m_view->setCaretStyle( KateRenderer::Underline, true );
-  m_viewInternal->repaint ();
+  m_viewInternal->update();
 }
 
 
@@ -425,7 +600,7 @@ void KateViInputModeManager::readSessionConfig( const KConfigGroup& config )
   // Reading marks
   QStringList marks = config.readEntry("ViMarks", QStringList() );
   for (int i = 0; i + 2 < marks.size(); i+=3) {
-      addMark(m_view->doc(),marks.at(i).at(0), KTextEditor::Cursor(marks.at(i+1).toInt(),marks.at(i+2).toInt()));
+      addMark(m_view->doc(),marks.at(i).at(0), Cursor(marks.at(i+1).toInt(),marks.at(i+2).toInt()));
   }
   syncViMarksAndBookmarks();
 }
@@ -474,7 +649,7 @@ void KateViInputModeManager::reset() {
         m_viVisualMode->reset();
 }
 
-void KateViInputModeManager::addJump(KTextEditor::Cursor cursor) {
+void KateViInputModeManager::addJump(Cursor cursor) {
    for (QList<KateViJump>::iterator iterator = jump_list->begin();
         iterator != jump_list->end();
         iterator ++){
@@ -493,7 +668,7 @@ void KateViInputModeManager::addJump(KTextEditor::Cursor cursor) {
 
 }
 
-KTextEditor::Cursor KateViInputModeManager::getNextJump(KTextEditor::Cursor cursor ) {
+Cursor KateViInputModeManager::getNextJump(Cursor cursor ) {
    if (current_jump != jump_list->end()) {
        KateViJump jump;
        if (current_jump + 1 != jump_list->end())
@@ -501,7 +676,7 @@ KTextEditor::Cursor KateViInputModeManager::getNextJump(KTextEditor::Cursor curs
        else
            jump = *(current_jump);
 
-       cursor = KTextEditor::Cursor(jump.line, jump.column);
+       cursor = Cursor(jump.line, jump.column);
    }
 
    // DEBUG
@@ -510,7 +685,7 @@ KTextEditor::Cursor KateViInputModeManager::getNextJump(KTextEditor::Cursor curs
    return cursor;
 }
 
-KTextEditor::Cursor KateViInputModeManager::getPrevJump(KTextEditor::Cursor cursor ) {
+Cursor KateViInputModeManager::getPrevJump(Cursor cursor ) {
    if (current_jump == jump_list->end()) {
        addJump(cursor);
        current_jump--;
@@ -519,7 +694,7 @@ KTextEditor::Cursor KateViInputModeManager::getPrevJump(KTextEditor::Cursor curs
    if (current_jump != jump_list->begin()) {
        KateViJump jump;
            jump = *(--current_jump);
-       cursor = KTextEditor::Cursor(jump.line, jump.column);
+       cursor = Cursor(jump.line, jump.column);
    }
 
    // DEBUG
@@ -543,14 +718,14 @@ void KateViInputModeManager::PrintJumpList(){
 
 }
 
-void KateViInputModeManager::addMark( KateDocument* doc, const QChar& mark, const KTextEditor::Cursor& pos,
+void KateViInputModeManager::addMark( KateDocument* doc, const QChar& mark, const Cursor& pos,
                                       const bool moveoninsert, const bool showmark )
 {
-  m_mark_set_inside_viinputmodemanager = true;
+  m_markSetInsideViInputModeManager = true;
   uint marktype = m_view->doc()->mark(pos.line());
 
   // delete old cursor if any
-  if (KTextEditor::MovingCursor *oldCursor = m_marks.value( mark )) {
+  if (MovingCursor *oldCursor = m_marks.value( mark )) {
 
     int  number_of_marks = 0;
 
@@ -560,22 +735,22 @@ void KateViInputModeManager::addMark( KateDocument* doc, const QChar& mark, cons
     }
 
     if (number_of_marks == 1 && pos.line() != oldCursor->line()) {
-      m_view->doc()->removeMark( oldCursor->line(), KTextEditor::MarkInterface::markType01 );
+      m_view->doc()->removeMark( oldCursor->line(), MarkInterface::markType01 );
     }
 
     delete oldCursor;
   }
 
-  KTextEditor::MovingCursor::InsertBehavior behavior = moveoninsert ? KTextEditor::MovingCursor::MoveOnInsert
-                                                                    : KTextEditor::MovingCursor::StayOnInsert;
+  MovingCursor::InsertBehavior behavior = moveoninsert ? MovingCursor::MoveOnInsert
+                                                                    : MovingCursor::StayOnInsert;
   // create and remember new one
   m_marks.insert( mark, doc->newMovingCursor( pos, behavior ) );
 
   // Showing what mark we set:
   if ( showmark && mark != '>' && mark != '<' && mark != '[' && mark != '.' && mark != ']') {
-    if( !marktype & KTextEditor::MarkInterface::markType01 ) {
+    if( !marktype & MarkInterface::markType01 ) {
       m_view->doc()->addMark( pos.line(),
-          KTextEditor::MarkInterface::markType01 );
+          MarkInterface::markType01 );
     }
 
     // only show message for active view
@@ -584,64 +759,66 @@ void KateViInputModeManager::addMark( KateDocument* doc, const QChar& mark, cons
     }
   }
 
-  m_mark_set_inside_viinputmodemanager = false;
+  m_markSetInsideViInputModeManager = false;
 }
 
-KTextEditor::Cursor KateViInputModeManager::getMarkPosition( const QChar& mark ) const
+Cursor KateViInputModeManager::getMarkPosition( const QChar& mark ) const
 {
   if ( m_marks.contains(mark) ) {
-    KTextEditor::MovingCursor* c = m_marks.value( mark );
-    return KTextEditor::Cursor( c->line(), c->column() );
+    MovingCursor* c = m_marks.value( mark );
+    return Cursor( c->line(), c->column() );
   } else {
-    return KTextEditor::Cursor::invalid();
+    return Cursor::invalid();
   }
 }
 
 
-void KateViInputModeManager::markChanged (KTextEditor::Document* doc,
-                                          KTextEditor::Mark mark,
-                                          KTextEditor::MarkInterface::MarkChangeAction action) {
+void KateViInputModeManager::markChanged (Document* doc,
+                                          Mark mark,
+                                          MarkInterface::MarkChangeAction action) {
 
-    Q_UNUSED( doc )
-  if (!m_mark_set_inside_viinputmodemanager) {
-    if (action == 1) {
-        foreach (QChar c, m_marks.keys()) {
-            if  ( m_marks.value(c)->line() == mark.line )
-                m_marks.remove(c);
-        }
-    } else if (action == 0) {
-      bool char_exist = false;
-      for( char c= 'a'; c <= 'z'; c++) {
-        if (!m_marks.value(c) && (mark.type == KTextEditor::MarkInterface::Bookmark)) {
-          addMark(m_view->doc(), c, KTextEditor::Cursor(mark.line,0));
-          char_exist = true;
-          break;
-        }
+  Q_UNUSED( doc )
+  if (mark.type != MarkInterface::Bookmark || m_markSetInsideViInputModeManager)
+  {
+    return;
+  }
+  if (action == MarkInterface::MarkRemoved) {
+      foreach (QChar markerChar, m_marks.keys()) {
+          if  ( m_marks.value(markerChar)->line() == mark.line )
+              m_marks.remove(markerChar);
+      }
+  } else if (action == MarkInterface::MarkAdded) {
+    bool freeMarkerCharFound = false;
+    for( char markerChar = 'a'; markerChar <= 'z'; markerChar++) {
+      if (!m_marks.value(markerChar)) {
+        addMark(m_view->doc(), markerChar, Cursor(mark.line, 0));
+        freeMarkerCharFound = true;
+        break;
       }
-      if (!char_exist)
-        m_viNormalMode->error(i18n ("There are no more chars for the next bookmark."));
     }
+    if (!freeMarkerCharFound)
+      m_viNormalMode->error(i18n ("There are no more chars for the next bookmark."));
   }
 }
 
 void KateViInputModeManager::syncViMarksAndBookmarks() {
-  const QHash<int, KTextEditor::Mark*> &m = m_view->doc()->marks();
+  const QHash<int, Mark*> &m = m_view->doc()->marks();
 
   //  Each bookmark should have a vi mark on the same line.
-  for (QHash<int, KTextEditor::Mark*>::const_iterator it = m.constBegin(); it != m.constEnd(); ++it)
+  for (QHash<int, Mark*>::const_iterator it = m.constBegin(); it != m.constEnd(); ++it)
   {
-    if (it.value()->type & KTextEditor::MarkInterface::markType01 ) {
-      bool there_is_vi_mark_for_this_line = false;
-      foreach( QChar key, m_marks.keys() ) {
-        if ( m_marks.value(key)->line() == it.value()->line ){
-          there_is_vi_mark_for_this_line = true;
+    if (it.value()->type & MarkInterface::markType01 ) {
+      bool thereIsViMarkForThisLine = false;
+      foreach( QChar markerChar, m_marks.keys() ) {
+        if ( m_marks.value(markerChar)->line() == it.value()->line ){
+          thereIsViMarkForThisLine = true;
           break;
         }
       }
-      if ( ! there_is_vi_mark_for_this_line ) {
-        for( char c= 'a'; c <= 'z'; c++ ) {
-          if ( ! m_marks.value(c) ) {
-            addMark( m_view->doc(), c, KTextEditor::Cursor( it.value()->line, 0 ) );
+      if ( !thereIsViMarkForThisLine ) {
+        for( char markerChar = 'a'; markerChar <= 'z'; markerChar++ ) {
+          if ( ! m_marks.value(markerChar) ) {
+            addMark( m_view->doc(), markerChar, Cursor( it.value()->line, 0 ) );
             break;
           }
         }
@@ -650,16 +827,16 @@ void KateViInputModeManager::syncViMarksAndBookmarks() {
   }
 
   // For each vi mark line should be bookmarked.
-  foreach( QChar key, m_marks.keys() ) {
-    bool there_is_kate_mark_for_this_line = false;
-    for (QHash<int, KTextEditor::Mark*>::const_iterator it = m.constBegin(); it != m.constEnd(); ++it) {
-      if (it.value()->type & KTextEditor::MarkInterface::markType01 ) {
-        if ( m_marks.value(key)->line() == it.value()->line ) {
-          there_is_kate_mark_for_this_line = true;
+  foreach( QChar markerChar, m_marks.keys() ) {
+    bool thereIsKateMarkForThisLine = false;
+    for (QHash<int, Mark*>::const_iterator it = m.constBegin(); it != m.constEnd(); ++it) {
+      if (it.value()->type & MarkInterface::markType01 ) {
+        if ( m_marks.value(markerChar)->line() == it.value()->line ) {
+          thereIsKateMarkForThisLine = true;
           break;
         }
-        if ( ! there_is_kate_mark_for_this_line) {
-          m_view->doc()->addMark( m_marks.value(key)->line(), KTextEditor::MarkInterface::markType01 );
+        if ( !thereIsKateMarkForThisLine) {
+          m_view->doc()->addMark( m_marks.value(markerChar)->line(), MarkInterface::markType01 );
         }
       }
     }
@@ -672,9 +849,9 @@ QString KateViInputModeManager::getMarksOnTheLine(int line) {
   QString res = "";
 
   if ( m_view->viInputMode() ) {
-    foreach (QChar c, m_marks.keys()) {
-      if  ( m_marks.value(c)->line() == line )
-        res += c + ":" + QString::number(m_marks.value(c)->column()) + " ";
+    foreach (QChar markerChar, m_marks.keys()) {
+      if  ( m_marks.value(markerChar)->line() == line )
+        res += markerChar + ":" + QString::number(m_marks.value(markerChar)->column()) + " ";
     }
   }
 
@@ -708,3 +885,32 @@ QString KateViInputModeManager::modeToString(ViMode mode)
 
   return modeStr;
 }
+
+KateViKeyMapper* KateViInputModeManager::keyMapper()
+{
+  return m_keyMapperStack.top().data();
+}
+
+KateViInputModeManager::Completion::Completion(const QString& completedText, bool removeTail, CompletionType completionType)
+    : m_completedText(completedText),
+      m_removeTail(removeTail),
+      m_completionType(completionType)
+{
+  if (m_completionType == FunctionWithArgs || m_completionType == FunctionWithoutArgs)
+  {
+    kDebug(13070) << "Completing a function while not removing tail currently unsupported; will remove tail instead";
+    m_removeTail = true;
+  }
+}
+QString KateViInputModeManager::Completion::completedText() const
+{
+  return m_completedText;
+}
+bool KateViInputModeManager::Completion::removeTail() const
+{
+  return m_removeTail;
+}
+KateViInputModeManager::Completion::CompletionType KateViInputModeManager::Completion::completionType() const
+{
+  return m_completionType;
+}
\ No newline at end of file
