@@ -21,6 +21,7 @@
 
 #include "katevikeyparser.h"
 #include <QStringList>
+#include <QKeyEvent>
 
 KateViKeyParser* KateViKeyParser::m_instance = NULL;
 
@@ -422,7 +423,7 @@ void KateViKeyParser::initKeyTables()
   m_nameToKeyCode->insert( QString( "d" ), 285 );
   m_nameToKeyCode->insert( QString( "e" ), 286 );
   m_nameToKeyCode->insert( QString( "f" ), 287 );
-  m_nameToKeyCode->insert( QString( "d" ), 288 );
+  m_nameToKeyCode->insert( QString( "g" ), 288 );
   m_nameToKeyCode->insert( QString( "h" ), 289 );
   m_nameToKeyCode->insert( QString( "i" ), 290 );
   m_nameToKeyCode->insert( QString( "j" ), 291 );
@@ -475,6 +476,17 @@ void KateViKeyParser::initKeyTables()
   m_nameToKeyCode->insert( QString( "?" ), 338 );
   m_nameToKeyCode->insert( QString( " " ), 339 );
   //m_nameToKeyCode->insert( QString( "<" ), 341 );
+  m_nameToKeyCode->insert( QString("0"), 340);
+  m_nameToKeyCode->insert( QString("1"), 341);
+  m_nameToKeyCode->insert( QString("2"), 342);
+  m_nameToKeyCode->insert( QString("3"), 343);
+  m_nameToKeyCode->insert( QString("4"), 344);
+  m_nameToKeyCode->insert( QString("5"), 345);
+  m_nameToKeyCode->insert( QString("6"), 346);
+  m_nameToKeyCode->insert( QString("7"), 347);
+  m_nameToKeyCode->insert( QString("8"), 348);
+  m_nameToKeyCode->insert( QString("9"), 349);
+
 
   for (QHash<QString, int>::const_iterator i = m_nameToKeyCode->constBegin();
       i != m_nameToKeyCode->constEnd(); ++i) {
@@ -581,9 +593,16 @@ const QString KateViKeyParser::encodeKeySequence( const QString &keys ) const
     }
     else {
       if ( c == '<' ) {
-        // if there's no closing '>', or if there is an opening '<' before the next '>', interpret as a literal '<'
+        // If there's no closing '>', or if there is an opening '<' before the next '>', interpret as a literal '<'
+        // If we are <space>, encode as a literal " ".
         QString rest = keys.mid( i );
-        if ( rest.indexOf( '>', 1 ) == -1 || ( rest.indexOf( '<', 1 ) < rest.indexOf( '>', 1 ) && rest.indexOf( '<', 1 ) != -1 ) ) {
+        if (rest.indexOf('>', 1) != -1 && rest.mid(1, rest.indexOf('>', 1) - 1) == "space")
+        {
+          encodedSequence.append(" ");
+          i += rest.indexOf('>', 1);
+          continue;
+        }
+        else if ( rest.indexOf( '>', 1 ) == -1 || ( rest.indexOf( '<', 1 ) < rest.indexOf( '>', 1 ) && rest.indexOf( '<', 1 ) != -1 ) ) {
           encodedSequence.append( c );
           continue;
         }
@@ -748,9 +767,13 @@ char KateViKeyParser::scanCodeToChar(quint32 code, Qt::KeyboardModifiers modifie
     return 0;
 }
 
-const QChar KateViKeyParser::KeyEventToQChar(int keyCode, const QString &text,
-    Qt::KeyboardModifiers mods, quint32 nativeScanCode)
+const QChar KateViKeyParser::KeyEventToQChar(const QKeyEvent& keyEvent)
 {
+  const int keyCode = keyEvent.key();
+  const QString &text = keyEvent.text();
+  const Qt::KeyboardModifiers mods = keyEvent.modifiers();
+  const quint32 nativeScanCode  = keyEvent.nativeScanCode();
+
   QChar key;
 
   if ( !text.isEmpty() ) {
@@ -764,8 +787,8 @@ const QChar KateViKeyParser::KeyEventToQChar(int keyCode, const QString &text,
       return key;
   }
 
-  if ( text.isEmpty() || ( text.length() == 1 && text.at(0) < 0x20 )
-      || ( mods != Qt::NoModifier && mods != Qt::ShiftModifier && mods != Qt::KeypadModifier ) ) {
+  if ( text.isEmpty() || ( text.length() == 1 && text.at(0) < 0x20 ) || keyCode == Qt::Key_Delete
+      || ( mods != Qt::NoModifier && mods != Qt::ShiftModifier && mods != Qt::KeypadModifier && mods != Qt::GroupSwitchModifier) ) {
     QString keyPress;
 
     keyPress.append( '<' );
